@@ -451,34 +451,6 @@ static struct pernet_operations kz_extension_net_ops = {
 	.exit_batch     = kz_extension_net_exit_batch,
 };
 
-static inline void kz_ct_dump_tuple_ip(const struct nf_conntrack_tuple *t, const u16 zone_id)
-{
-	pr_err_ratelimited("existing item in kzorp hash with the same tuple %p: %u %pI4:%hu -> %pI4:%hu\n",
-			   t, t->dst.protonum,
-			   &t->src.u3.ip, ntohs(t->src.u.all),
-			   &t->dst.u3.ip, ntohs(t->dst.u.all));
-}
-
-static inline void kz_ct_dump_tuple_ipv6(const struct nf_conntrack_tuple *t, const u16 zone_id)
-{
-	pr_err_ratelimited("existing item in kzorp hash with the same tuple %p: %u %pI6 %hu -> %pI6 %hu\n",
-			   t, t->dst.protonum,
-			   t->src.u3.all, ntohs(t->src.u.all),
-			   t->dst.u3.all, ntohs(t->dst.u.all));
-}
-
-static inline void kz_ct_dump_tuple(const struct nf_conntrack_tuple *t, const u16 zone_id)
-{
-	switch (t->src.l3num) {
-	case AF_INET:
-		kz_ct_dump_tuple_ip(t, zone_id);
-		break;
-	case AF_INET6:
-		kz_ct_dump_tuple_ipv6(t, zone_id);
-		break;
-	}
-}
-
 /* deallocate entries in the hashtable */
 static void clean_hash(void)
 {
@@ -488,7 +460,10 @@ static void clean_hash(void)
 		while (!hlist_nulls_empty(&kz_hash[i])) {
 			struct kz_extension *kzorp = kz_extension_get_from_node(kz_hash[i].first);
 			kz_extension_remove_from_cache(kzorp);
-			kz_ct_dump_tuple(&kzorp->tuple_orig, kzorp->zone_id);
+			kz_log_with_dumped_tuple_err_ratelimited(
+				"existing item in kzorp hash with the same tuple",
+				&kzorp->tuple_orig, kzorp->zone_id
+			);
 			kz_extension_put(kzorp);
 		}
 	}
